/*  
	bootparam_get_router.c
	jelinekr@ms.com		11/15/04
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>

#define ALL_ROUTERS	"224.0.0.2"
#define PACKETLEN	8
#define	MAXPACKETLEN	4096
#define MAXNUMROUTERS	1024
#define MAXTIMEOUTCNT	1
#define MAXBADPACKETS	1

#define	TIMEOUT		300000

extern int		debug;

u_long			get_router(struct in_addr);
static int		send_pack(int, struct sockaddr_in *);
static u_long		parse_pack(char *, int, struct sockaddr_in *);
static u_short		in_cksum(u_short *, int);
static int		ulong_compare(const void *, const void *);
static int		support_multicast(void);

#ifdef  __STDC__
#include <stdarg.h>

void msgout(char *fmt, ...);

#else
#include <varargs.h>

void msgout(char *fmt, va_alist va_dcl);

#endif /* __STDC__ */


/*
 * Find a router by sending ICMP_ECHO to ALL_ROUTERS multicast group
 */
u_long
get_router(src_addr)
	struct in_addr		src_addr;
{
	struct sockaddr_in	*to;
	struct sockaddr_in	where_to;
	struct sockaddr_in	if_addr;
	struct protoent		*proto;
	char			*dest_addr;
	int                     sock;


	if (! support_multicast()) {
		msgout("Multicast not supported");
		return((u_long)0);
	}

	dest_addr = ALL_ROUTERS;
	bcopy(&src_addr.s_addr, &if_addr.sin_addr.s_addr, sizeof(struct in_addr));
	if_addr.sin_family = AF_INET;

	if (debug)
		msgout("Source address %s", inet_ntoa(if_addr.sin_addr));


	bzero((char *)&where_to, sizeof(struct sockaddr_in));
	to = (struct sockaddr_in *)&where_to;
	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(dest_addr);

	if (to->sin_addr.s_addr == (u_long)-1) {
		msgout("Bad destination address %s", dest_addr);
		return((u_long)0);
	}

	if ((proto = getprotobyname("icmp")) == NULL) {
		msgout("Unknown protocol icmp");
		return((u_long)0);
	}

	if ((sock = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
		perror("socket");
		return((u_long)0);
	}

	if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_IF,
			(char *)&if_addr.sin_addr, sizeof (if_addr.sin_addr)) < 0) {
		msgout("Cannot send multicast packet over interface %s", 
			inet_ntoa(if_addr.sin_addr));
		close(sock);
		return((u_long)0);
	}

	if (! send_pack(sock, &where_to)) {
		close(sock);
		return((u_long)0);
	}

	{
		struct sockaddr_in	from;
		u_int			fromlen = sizeof(from);
		char			buf[MAXPACKETLEN];
		int                     nbytes, buflen = sizeof(buf);
		u_long                  routers[MAXNUMROUTERS];
		u_long			router = 0;
		int			i, nfd, cnt = 0;
		int			tocnt = 0;
		int			bpcnt = 0;

		struct timeval		timeout;
		fd_set			readset;

		timeout.tv_sec  = 0;
		timeout.tv_usec = TIMEOUT;

		FD_ZERO(&readset);
		FD_SET(sock, &readset);

		memset (routers, '\0', sizeof(routers));

		while (1) {

			nfd = select(sock+1, &readset, (fd_set *)0 , NULL, &timeout);
  			if (nfd < 0) {
				if (errno == EINTR)
					continue;
				perror("select");
				close(sock);
				return((u_long)0);
			} else if (nfd == 0) {
				tocnt++;

				if (debug)
					msgout("Timed out in recvfrom() (%d)", tocnt);

				if (tocnt > MAXTIMEOUTCNT)
					break;
				else
					continue;

			}

			nbytes = recvfrom(sock, (char *)buf, buflen, 0, 
					(struct sockaddr *)&from, &fromlen);

			if (nbytes > 0) {
				if (debug)
					msgout("Read %d bytes (%d)", nbytes, cnt);

				if ((router = (u_long)parse_pack((char *)buf, nbytes, &from)) != 0) {
					if (cnt < MAXNUMROUTERS) {
						/* convert to host byte order for sorting */
						routers[cnt] = ntohl(router);
						cnt++;
					} else
						break;
				} else {
					bpcnt++;

					if (debug)
						msgout("Bad packet from recvfrom() (%d)", bpcnt);

					if (bpcnt > MAXBADPACKETS)
						break;
					else
						continue;
				}
			}
		}

		close (sock);

		qsort((void *)routers, cnt, sizeof (u_long), ulong_compare);

		for (i = 0; i < cnt; i++) {
			if (debug)
				msgout("Routers %x (%d)", routers[i], i);
		}
		
		/* return the one with lowest IP in network byte order */
		return ((u_long)htonl(routers[0]));
	}
}

/*
 * Send an ICMP packet
 */
static int
send_pack(sock, dst)
	int			sock;
	struct sockaddr_in	*dst;
{
	char			buf[PACKETLEN];
	int			nbytes, buflen = sizeof(buf);
	struct icmp             *icp = (struct icmp *)buf;

	icp->icmp_type = ICMP_ECHO; /* ICMP_ROUTERSOLICIT */
	icp->icmp_code = 0;
	icp->icmp_void = 0;
	icp->icmp_seq = 0;
	icp->icmp_id = getpid();
	icp->icmp_cksum = 0;


	icp->icmp_cksum = in_cksum((u_short *)icp, buflen);

	nbytes = sendto(sock, (char *)buf, buflen, 0, 
		(struct sockaddr *)dst, sizeof(struct sockaddr));

	if (nbytes < 0) {
		perror ("sendto");
		return(0);
	}

	if (nbytes != buflen) {
		msgout("Sent %d bytes, expected %d", nbytes, buflen);
		return(0);
	}

	return(nbytes);
}

/*
 * Process the received ICMP packet
 */

static u_long
parse_pack(buf, nbytes, from)
	char			*buf;
	int			nbytes;
	struct sockaddr_in	*from;
{
	int			hlen;
	struct icmp		*icp;
	struct ip		*ip;
	struct in_addr		*router;


	ip = (struct ip *)buf;
	hlen = ip->ip_hl << 2;

	if (nbytes < hlen + ICMP_MINLEN) {
		msgout("Packet too short %d bytes from %s", nbytes,
			inet_ntoa(*(struct in_addr *)&from->sin_addr.s_addr));
		return((u_long)0);
	}

	nbytes -= hlen;
	icp = (struct icmp *)(buf + hlen);

	if (icp->icmp_type == ICMP_ECHOREPLY) { /* ICMP_ROUTERSOLICIT */

		if (nbytes != PACKETLEN) {
			msgout("Received nbytes %d, expected %d", nbytes, PACKETLEN);
			return((u_long)0);
		}

		if (icp->icmp_seq != 0) {
			msgout("Wrong sequence %d", icp->icmp_seq);
			return((u_long)0);
		}

		if (icp->icmp_id != getpid()) {
			msgout("Wrong id %d, expected %d", icp->icmp_id, getpid());
			return((u_long)0);
		}

		router = (struct in_addr *)&from->sin_addr;

		if (debug)
			msgout("Received reply from %s", inet_ntoa(*router));

		return((u_long)router->s_addr);
	}

	return((u_long)0);
}

/*
 * in_cksum --Checksum routine for Internet Protocol family headers (C Version)
 *
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
static u_short
in_cksum(addr, len)
	u_short			*addr;
	int			len;
{
	register int		nleft = len;
	register u_short	*w = addr;
	register int		sum = 0;
	u_short			answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(unsigned char *)(&answer) = *(unsigned char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

static int
support_multicast(void)
{
        int			sock;
        unsigned char		ttl = 1;

        if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
                perror("socket");
                return(0);
        }

        if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, sizeof (ttl)) < 0) {
		perror("setsockopt");
                close(sock);
                return(0);
        }
        close(sock);
        return(1);
}

static int
ulong_compare(p1, p2)
	const void		*p1;
	const void		*p2;
{
	return(*(u_long *)p1 - *(u_long *)p2);
}
